/*
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2010,2012. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 */
define([
"dojo/_base/declare",
"dojo/_base/sniff",
"dojo/_base/window",
"dojo/_base/connect",
"dojo/dom-attr",
"dojox/gfx/matrix",
"./Viewport",
"./ScrollbarHudComponent",
"../util/HandleSet",
"../util/Geometry",
"../util/ErrorReporter"
], function(
declare,
has,
window,
connect,
doma,
matrix,
Viewport,
ScrollbarHudComponent,
HandleSet,
g,
R
){

var Size = g.Size;

/*=====
var Viewport = ibm_ilog.diagram.widget.Viewport;
var HandleSet = ibm_ilog.diagram.util.HandleSet;
=====*/

var GfxViewport =
declare('ibm_ilog.diagram.widget.GfxViewport', [Viewport, HandleSet], {
	//
	//	summary:
	//		This is a Viewport class based on an HTML <div> which overflows with scrollbars.
	//
	//	description:
	//
	//		DivViewport depends on a div declared at Diagram.html with attachPoint=scrollbarsNode.
	//		This div encloses the diagram Surface. 
	//		
	//		The diagram Surface is extended to enclose the whole diagram tightly. The 
	//		div is constrained to the desired client size using its width and height 
	//		properties. The view rectangle is then associated with the div scroll 
	//		positions, scrollTop and scrollLeft. Setting the view rectangle then 
	//		updates scrollTop and scrollLeft.
	//
	//		The zoom level is derived from the view rectangle, but is mapped entirely 
	//		different than its position. Instead, the zoom controls a scale transformation 
	//		of the entire Diagram. If the view rectangle is small, the Diagram is zoomed 
	//		in to fit the requested portion into the viewport window. With this also the 
	//		Surface is scaled, as it is required that the entire Graph fits the surface 
	//		exactly so that the scrollTop/scrollLeft properties work.
	//		
	//	note:
	//		the zooming service required from the Diagram object should be at a UI 
	//		level and should not affect, for example, the bouding box of the the Graph 
	//		objects. One option is to define an intermediate node between the Surface root 
	//		and the Graph node, which controls this transformation.
	//
		
	_hScrollbarVisible: true,
	
	_vScrollbarVisible: true,

	templateString: function() {
		return "<div style='overflow:hidden; position:relative'>"+
		        	"<div dojoAttachPoint='canvasNode' style='z-index:1; position:absolute; top:0px; left:0px'></div>"+
		        "</div>";
	},
	
	//
	//	_updatingScrollbars:
	//		true when scrollbar position is being updated and we will ignore related events
	//		useful in chrome since it reports events immediately.
	//
	_updatingScrollbars: false,

	getHudSize: function() {
		var hsz = this.getClientSize();
//		var sbw = 16;//this.getScrollbarWidth();
//		if(!this._hScrollbarVisible) {
//			hsz.height+=sbw;
//		}
//		if(!this._vScrollbarVisible) {
//			hsz.width+=sbw;
//		}
		return hsz;
	},
	
	updateScrollbarVisibilityFlags: function() {
  	if(this._hScrollbarVisible != (this._viewRect.width<this._contentRect.width)) {
  		this._hScrollbarVisible = (this._viewRect.width<this._contentRect.width);
  		this._needsHudRelayout = true;
  	}
  	if(this._vScrollbarVisible != (this._viewRect.height<this._contentRect.height)) {
  		this._vScrollbarVisible = (this._viewRect.height<this._contentRect.height);
  		this._needsHudRelayout = true;
  	}
	},

	_onContentBoundingBoxChanged: function(){
		//
		//	summary:
		//		The function is executed when the content of the content has changed.
		//
		//	returns:
		//		True if the change affected content bounding box.
		//
		//	tags:
		//		private
		//
		this.inherited(arguments)
	},

	initialize: function( /*ibm_ilog.diagram.widget.Diagram*/ diagram ) {
		//
		//	summary:
		//		Initializes a Viewport for collaborating with the diagram.
		//
		//	diagram: ibm_ilog.diagram.widget.Diagram
		//		The diagram
		//
//		this._scrollbars = diagram.scrollbarsNode;		
//		this._scrollbarsContent = diagram.scrollbarsContentNode;
		
		this.inherited(arguments);

		this._surface.whenLoaded(this, function () {
			new ScrollbarHudComponent(this,true);
			new ScrollbarHudComponent(this,false);
		});

		//this._connect("onscroll",this._scrollbars,"onscroll","_onScroll");
		
	},

//	_onScroll: function(e) {
//		if(!this._updatingScrollbars) {
//			var p = Point(this._scrollbars.scrollLeft,this._scrollbars.scrollTop);
//			
//			// The following condition traps bouncing events generated by 
//			// the receiver updating the scrollbar programatically
//			
//			if(!g.equalPoint(this._scrollCache,p)) {
//				var q = this._pointScrollToModel(p);
//				this.setViewRect(q,{instant:true});
//			}
//		}
//	},	

	setClientSize: function( /*Size*/ sz ){
		//
		//	summary:
		//		Sets the client Size
		//
		this.setSize({width:sz.width,height:sz.height});
	},
	
	setSize: function( /*Size*/ sz ) {
		//
		//	summary:
		//		Sets the Viewport Size
		//
		var style = {style:{width:sz.width+"px",height:sz.height+"px"}};

		doma.set(this._diagram.domNode,style);
		//dojo.attr(this._scrollbars,style);
		this._updateClientSize(Size(sz.width,sz.height));
		this._updateSurface();
	},
	

	_updateClientSize: function( /*Size*/ sz ) {
		//
		//	summary:
		//		Updates viewportRect from the div attributes
		//
		//	tags:
		//		private
		//
		var vpH = sz.height ;//|| this._scrollbars.clientHeight;
		var vpW = sz.width ;//|| this._scrollbars.clientWidth;
		
		var newClientSize = { width:vpW, height:vpH };


		if(!g.equalRect(this._clientSize,newClientSize)) {

			this._clientSize = newClientSize;
			
			this._computeContentRect();

			switch(this._resizeMode) {
			case 1: // keep zoom level
				this.setViewRect({
					x:this._viewRect.x,
					y:this._viewRect.y,
					width:(newClientSize.width/this.getZoom())},{instant:true});
				break;
			case 2:
				this._onViewRectConstraintsChanged();
				break;
			default:
				R.error("InvalidViewportResizeMode",this._resizeMode)
			}
			
			this.onClientSizeChanged();

		}
	},

	_pointContentToScroll: function(p) {
		//
		//	summary:
		//		Transforms a point from content coordinates to UI coordinates.
		//
		return g.subPoint(this._viewportMatrix.transformPoint(p),this._transformedContentRect);
	},
	
	_pointScrollToModel: function(p) {
		//
		//	summary:
		//		Transforms a point from UI coordinates to content coordinates.
		//
		return this._viewportMatrixInverse.transformPoint(g.addPoint(this._transformedContentRect,p));
	},	

	
	onViewRectChanged: function onViewRectChanged(){
		this.updateScrollbarVisibilityFlags();	
		this.inherited(arguments);
	},

	_applyViewRect: function( /*Boolean*/ sizeChanged, /*Boolean*/ originChanged ){
		//
		//	summary:
		//		Updates the surface size and _transformedContentRect based on the _viewRect and
		//		_contentRect. Called after a change in zoom or Graph content.
		//
		//	tags:
		//		private
		//
		
		this.inherited(arguments);
		
		// create transformation, apply if to Graph, and compute new transformedContentRect
		var zt = matrix.scale(this._zoom);
		var tt = matrix.translate(g.negPoint(this._viewRect)); // uses rect origin
		var t  = matrix.multiply(zt,tt);

		this._transformedContentRect = t.transformRectangle(this._contentRect);
		
		var scrollContentSize = g.floorSize(g.growRect(this._transformedContentRect,this.getClientSize()));
		      // this._transformedContentRect;
			 	  // [AV] replace by this go back to tight surface mode

		this._viewportMatrix = t;
		this._viewportMatrixInverse = t.inverse();

		this._setViewportTransform(t);
	},
	
	_getSurfaceSize: function () {
		return this._clientSize;
	}

});

return GfxViewport;

});
